/*NSRCOPYRIGHT
	Copyright (C) 1999-2011 University of Washington
	Developed by the National Simulation Resource
	Department of Bioengineering,  Box 355061
	University of Washington, Seattle, WA 98195-5061.
	Dr. J. B. Bassingthwaighte, Director
END_NSRCOPYRIGHT*/

// look up ontology terms description for database

package JSim.db;

import JSim.util.*;
import java.io.*;
import java.net.*;
import java.util.*;
import org.w3c.dom.*;

import uk.ac.ebi.biomodels.ws.*;
import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import javax.xml.namespace.QName;

public class DBOntLookup {
	Hashtable<String, String> bioportalVersionIDs;
	BioModelsWSClient bioClient;
	Service ecService;
	URL ecEndpoint;
	QName ecQname;
	
	// constructor
	public DBOntLookup() {
	    bioportalVersionIDs = new Hashtable<String, String>();
	    bioportalVersionIDs.put("obo.chebi", "50031");
	    bioportalVersionIDs.put("obo.go", "50352");
	    bioportalVersionIDs.put("obo.bto", "49596");
	    bioportalVersionIDs.put("obo.pato", "50065");
	    bioportalVersionIDs.put("obo.psi-mod", "18960");
	    bioportalVersionIDs.put("obo.sbo", "45638");
	    bioportalVersionIDs.put("obo.clo", "50349");
	    bioportalVersionIDs.put("taxonomy", "47845");
	}

	// process terms in file
	public void process(File f) throws Exception {
	    String ftext = UtilIO.readText(f);
	    StringList list = new StringList(ftext, "\n");
	    for (int i=0; i<list.size(); i++) {
	        String line = list.str(i).trim();
	    	StringTokenizer stok = new StringTokenizer(line, " \t");
		if (stok.countTokens() != 2) {
		    System.err.println("BAD-INPUT: " + line);
		    continue;
		}
		String ont = stok.nextToken();
		String tag = stok.nextToken();
		try {
		    String desc = getDesc(ont, tag);
		    if (desc == null) {
		    	System.err.println("UNKNOWN-ONTOLOGY: " + line);
		    	continue;
		    }
		    System.out.println(ont + "\t" + tag + "\t\"" + desc + "\"");
		} catch (Exception e) {
		    System.err.println("LOOKUP-FAILED: " + line);
		}
	    }
	}
	
	
	// get description for ontology/term
	public String getDesc(String ontology, String tag) 
	throws Exception {
	    if (ontology.startsWith("kegg")) 
	    	return getKeggDesc(ontology, tag);
	    if (ontology.equals("uniprot"))
	    	return getUniprotDesc(tag);
	    if (ontology.equals("interpro"))
	    	return getInterproDesc(tag);
	    if (ontology.equals("biomodels.db"))
	    	return getBiomodelsDesc(tag);
	    if (ontology.equals("ec-code"))
	    	return getEcCodeDesc(tag);
	    if (ontology.equals("reactome"))
	    	return getReactomeDesc(tag);
	    String versID = bioportalVersionIDs.get(ontology);
	    if (versID != null) {
	    	if (ontology.equals("taxonomy"))
		    tag = "obo:NCBITaxon_" + tag;
		if (ontology.equals("obo.sbo"))
		    tag = tag.toUpperCase();
	    	return getBioportalTag(versID, tag,
		    new String[] { "data", "classBean", "label" });
	    }
	    if (ontology.equals("pubmed"))
	    	return getPubMedDesc(tag);
	    return null;
	}
	
	// KEGG REST service
	private String getKeggDesc(String ontology, String tag) 
	throws Exception {
	    URL url = new URL("http://rest.kegg.jp/get/" + tag);
	    System.err.println("Grabbing " + url);
	    String text = UtilIO.readText(url);
//	    System.err.println(text);
	    StringList lines = new StringList(text, "\n");
	    for (int i=0; i<lines.size(); i++) {
	    	String line = lines.str(i);
		if (! line.startsWith("NAME")) continue;
		line = line.substring(4).trim();
		if (line.endsWith(";")) 
		    line = line.substring(0, line.length()-1);
		return line;
	    }
	    return null;
	}
	
	// Uniprot REST service
	private String getUniprotDesc(String tag) 
	throws Exception {
	    URL url = new URL("http://www.uniprot.org/uniprot/" + tag + ".xml");
	    return getNestedXMLTag(url, 
	    	new String[] { "entry", "protein", "recommendedName", "fullName" });
	}

	// Interpro REST service
	private String getInterproDesc(String tag) 
	throws Exception {
	    URL url = new URL(
	    	"http://www.ebi.ac.uk/Tools/dbfetch/dbfetch/interpro/" 
		+ tag + "/interproxml");
	    return getNestedXMLTag(url, 
	    	new String[] { "interpro", "name" });
	}

	// get bioportal label
	private String getBioportalTag(String versionID, 
	String tag, String[] tags) throws Exception {
	    URL url = new URL(
	    	"http://rest.bioontology.org/bioportal/concepts/" 
		+ versionID + "?conceptid=" + tag + 
		"&apikey=a803f56b-fc4d-4f6e-a452-cf75b624fb9a");
	    return getNestedXMLTag(url, tags);
	} 

	// get PubMed desc
	private String getPubMedDesc(String tag) throws Exception {
	    URL url = new URL(
	        "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi" 
		+ "?db=pubmed&id=\ufffd" + tag
		+ "&rettype=xml&retmode=text");
	    String[] tags = new String[] {
	    	"PubmedArticle",  
		"MedlineCitation",
		"Article",
		"ArticleTitle"
	    };
	    return getNestedXMLTag(url, tags);
	}

	// get Reactome desc
	// URL text grabbed in Java is not same as same grabbed in browser
	//   worked around for now
	// REST service in development
	//   Some Biomodels tags are from non-current release,
	//     and are not (yet?) supported
	private String getReactomeDesc(String tag) throws Exception {
	    int inx = tag.indexOf(".");
	    if (inx > 0) tag = tag.substring(0,inx);
	    URL url = new URL(
		"http://reactomews.oicr.on.ca:8080/ReactomeRESTfulAPI"
		+ "/RESTfulWS/queryById/DatabaseObject/" + tag);
	    String text = UtilIO.readText(url);
	    StringList list = new StringList(text, "\n");
	    for (int i=0; i<list.size(); i++) {
	    	String line = list.str(i);
		// System.err.println("line=" + line);
	    	StringTokenizer ltok = new StringTokenizer(line, "\"");
		if (ltok.countTokens() < 4) continue;
		String s1 = ltok.nextToken();
		String s2 = ltok.nextToken();
		String s3 = ltok.nextToken();
		String s4 = ltok.nextToken();
		if (s2.equals("displayName")) return s4;
	    }
	    return null;
	}

	// get nested text from XML text
	private String getNestedXMLTag(URL url, String[] tags) 
	throws Exception {
  	    System.err.println("Grabbing " + url);
	    String text = UtilIO.readText(url);
	    // System.err.println(text);
	    Document doc = UtilXML.parse(text);
	    Element e = doc.getDocumentElement();
	    e = UtilXML.getNestedElement(e, tags);
//System.err.println("e=" + e);
	    String name = UtilXML.getText(e);
//System.err.println("name=" + name);
	    name = name.trim();
	    return (name.length() > 0) ? name : null;
	}
	    

	// get Biomodels via SOAP
	private String getBiomodelsDesc(String tag) 
	throws Exception {
	    if (bioClient == null)
	    	bioClient = new BioModelsWSClient();
	    return bioClient.getModelNameById(tag);
	}

	// get Enzyme Commision recommended name via SOAP
	//   see http://www.brenda-enzymes.org/
	private String getEcCodeDesc(String tag) 
	throws Exception {
	    if (ecService == null) {
		ecService = new Service();
		ecEndpoint = new URL(
		    "http://www.brenda-enzymes.org/soap2/brenda_server.php");
		ecQname = new QName(
		    "http://soapinterop.org/", "getRecommendedName");
	    }
	    Call call = (Call) ecService.createCall();
	    call.setTargetEndpointAddress(ecEndpoint);
	    call.setOperationName(ecQname);
	    String qs = "ecNumber*" + tag;
	    Object[] qargs = new Object[] { qs };
	    String result = (String) call.invoke(qargs);
	    if (result == null) return null;
    	    // System.err.println("EC result=" + result);
	    StringTokenizer stok = new StringTokenizer(result, "#");
	    while (stok.hasMoreTokens()) {
	    	String tok = stok.nextToken();
		if (! tok.startsWith("recommendedName*")) continue;
		int inx = tok.indexOf("*");
		return tok.substring(inx+1);
	    }
	    return null;
	}

	// mainline process files
	public static final void main(String[] args) throws Exception {
	    DBOntLookup lookup = new DBOntLookup();
	    for (int i=0; i<args.length; i++) {
	    	File f = new File(args[i]);
		lookup.process(f);
	    }
	}
}	
