/*NSRCOPYRIGHT
	Copyright (C) 1999-2011 University of Washington
	Developed by the National Simulation Resource
	Department of Bioengineering,  Box 355061
	University of Washington, Seattle, WA 98195-5061.
	Dr. J. B. Bassingthwaighte, Director
END_NSRCOPYRIGHT*/

// generate attractive tic intervals and labels

package JSim.gui.gsgraph;

import JSim.util.*;

public class GSTics {

        // instance fields
	private boolean log;  // log scale?
        private double axismin;  // min value on axis
        private double axismax;  // max value on axis
        private int deltax;    // axis delta index (see delta(int))

	// constructor 
        public GSTics(
        double min,     // minimum data or user range,  or NaN if no data 
        double max,     // maximum data or user range,  or NaN if no data
        boolean l,    // true if log scale, otherwise linear
        boolean auto,    // if true, min and max come from data, 
                              // otherwise from user
	int nchars) // #chars available for tics,  or 0 if not applicable
        {
	    // HACK!!! prevent GSTics infinite loop
	    if (Double.isInfinite(min)) min=Double.NaN;
	    if (Double.isInfinite(max)) max=Double.NaN;

	    // ensure axismin <= axismax
	    axismin = min;
	    axismax = max;
            if (axismax < axismin) {
                double temp = axismax;
                axismax = axismin;
                axismin = temp;
            }

	    log = l;
	    if (axismin <= 0) log = false;

	    if (log) 
		doLog();
	    else
		doLinear(auto, nchars);
	}

	// log tic calculation
	private void doLog() {

            // ensure axismin < axismax
            if (axismin == axismax) axismax *= 100;

	    // calculate deltax
	    deltax = (int) (3*Math.log(axismin)/LOG10);
	}

	// linear tic calculation
	private void doLinear(boolean auto, int nchars) {

            // ensure axismin < axismax
            if (axismin == axismax) {
		double d = 
		    (axismin==0) ? 1d : Math.abs(axismin);
                axismin -= d;
                axismax += d;
            }

	    // protect against too small a range
	    double range = axismax - axismin;
	    if (axismin/range > THRESH || axismax/range > THRESH) {
		double center = (axismin==0) ? axismax : axismin;
		axismin -= center/THRESH;
		axismax += center/THRESH;
	    }

	    // decrement so deltax has at least 2 tics
	    range = axismax - axismin;
	    deltax = (int) (3*Math.log(range)/LOG10);
	    while (ntics(axismin, axismax, delta(deltax)) < 2)
		deltax--;

	    // auto axis range?
	    if (auto) {
		double d = delta(deltax-2);
		axismin = d * Math.floor(axismin/d);
		axismax = d * Math.ceil(axismax/d);
	    	while (ntics(axismin, axismax, delta(deltax+1)) >= 2)
		    deltax++;
	    } 

/*	    // show deltax alternatives (debug only)
	    for (int i=0; i<3; i++) {
		double delta = delta(deltax-i);
		int prec = ticPrecision(axismin, axismax, delta);
		int n = ntics(axismin, axismax, delta);
		double[] vals = ticValues(axismin, axismax, delta);
		int chars = ticMaxChars(vals, prec);
		StringList labs = ticLabels(vals, prec);
	    	System.err.println(
		    "try deltax=" + (deltax-i) + " delta=" + delta +
		    " ntics=" + ntics(axismin, axismax, delta) +
		    " prec=" + prec + " chars=" + chars + 
		    " tics=" + labs);
	    }
*/

	    // decrease deltax so long as space avail and #tics < 10
	    while (true) {
		double delta = delta(deltax-1);
		int prec = ticPrecision(axismin, axismax, delta);
		int n = ntics(axismin, axismax, delta);
		int maxTics = log ? 9 : 10;
 		if (n<2 || n>=maxTics) break; // <2 for pathological 

		// test space avail
		if (nchars > 0) {
		    int tspace = (int) Math.floor(nchars*(delta/(axismax-axismin)));
		    double[] vals = ticValues(axismin, axismax, delta);
		    int tchars = ticMaxChars(vals, prec);
		    if (tchars >= tspace) break;
		}

		// decrement deltax and continue
		deltax--;
	    }

	}

	// query
	public boolean log() { return log; }
	public double axismin() { return axismin; }
	public double axismax() { return axismax; }
	public double axisrange() { return axismax-axismin; }
	public double delta() { return delta(deltax); }
	public int ntics() {
	    if (log) return 0;
	    return ntics(axismin, axismax, delta()); 
	}
	public double[] ticValues() {
	    if (log) return null;
	    return ticValues(axismin, axismax, delta());
	}
	public String[] ticLabels() {
	    if (log) return null;
	    int prec = ticPrecision(axismin, axismax, delta());
	    return ticLabels(ticValues(), prec).array();
	}
	public String prettyQuery(double v) {
	    int prec = log ? 6 : ticPrecision(axismin, axismax, delta()) + 2;
	    return format(v, prec);
	}
	public String toString() {
	    String s = "{" + prettyQuery(axismin) +
		" " + prettyQuery(axismax) +
		" " + prettyQuery(delta()) +
		"} ";
	    if (log) return s + "log";

 	    int prec = ticPrecision(axismin, axismax, delta());
	    StringList labs = ticLabels(ticValues(), prec);
	    return s + labs;
	}

	//// STATIC VARS & METHODS

	// constant
	private static final double THRESH = 1e7;
	private static final double LOG10 = Math.log(10);

	// numeric formatting methods
	private static PrettyFormat prettyFormat[] = new PrettyFormat[32];
	private static PrettyFormat prettyFormat(int i) {
	   if (i>=prettyFormat.length) return new PrettyFormat(i);
	   if (prettyFormat[i] == null)
		prettyFormat[i] = new PrettyFormat(i);
	   return prettyFormat[i];
	}
	private static String format(double v, int i) {
	    return prettyFormat(i).format(v);
	}
	
	// linear delta for inx
	private static double delta(int i) {

	    // calc power10 and remainder
	    int p, r;
	    if (i>=0) {
		p = i/3;
		r = i%3;
	    } else {
		p = (i-2)/3;
		r = (-i)%3;
		if (r>0) r = 3-r;
	    }

	    // calculate d=10^p but more stable than Math.pow(10,p)
	    double d = 1;
	    if (p>0)
		for (int j=0; j<p; j++) d *= 10;
	    else if (p<0)
		for (int j=p; j<0; j++) d /= 10;
	    
	    // return delta
	    switch (r) {
	    case 1: return d*2;
	    case 2: return d*5;
	    default: return d;
	    }
	}

	// # tics for min/max/delta
	private static int ntics(double min, double max, double delta) {
	    int lox = (int) Math.ceil(min/delta - 1/THRESH); 
	    int hix = (int) Math.floor(max/delta + 1/THRESH);
	    return hix-lox+1;
	}

	// tic values for min/max/delta
	private static double[] ticValues(double min, double max, double delta) {
	    int n = ntics(min, max, delta);
	    double[] v = new double[n];
	    int minx = (int) Math.ceil(min/delta - 1/THRESH);
	    for (int i=0; i<n; i++) 
		v[i] = delta*(minx+i);
	    return v;
	}

	// tic labels
	private static StringList ticLabels(double[] v, int prec) {
	    StringList labs = new StringList(v.length);
	    for (int i=0; i<v.length; i++)
	    	labs.add(format(v[i], prec));
	    return labs;
	}

	// max # chars for tic label rendering
	private static int ticMaxChars(double[] v, int prec) {
	    int ct = 0; 
	    for (int i=0; i<v.length; i++) {
		String s = format(v[i], prec);
		int j = s.length();
		if (j>ct) ct = j;
	    }
	    return ct+4; // always leave 2 spaces on either side 
	}

	// precision required for tic rendering min/max/delta
   	private static int ticPrecision(double min, double max, 
	double delta) {
	    double mintic = Math.abs(delta * Math.ceil(min/delta - 1/THRESH));
	    double maxtic = Math.abs(delta * Math.floor(max/delta + 1/THRESH));
	    double tic = Math.max(mintic, maxtic);
	    int prec = (int) Math.ceil(Math.log(tic/delta)/LOG10);
	    prec++; // make up for PrettyFormat irregularities
	    if (prec < 3) prec = 3; // sanity check
	    return prec;
	}

	    
	// test program
	public static void main(String[] args) throws Exception {
            if (args.length != 5) 
		throw new Exception(
                    "Usage: GSTics min max log autoscale nchars");

            int i = 0;
            double min = Util.toDouble(args[i++]);
            double max = Util.toDouble(args[i++]);

            boolean log = false;
            switch (args[i++].charAt(0)) {
            case 't':
            case 'T':        
                log = true;
                break;
            case 'f':
            case 'F':        
                log = false;
                break;
            default:
                throw new Exception(
                        "Parameter \"log\" must be true or false");
            }

            boolean autoscale = true;
            switch (args[i++].charAt(0)) {
            case 't':
            case 'T':        
                autoscale = true;
                break;
            case 'f':
            case 'F':        
                autoscale = false;
                break;
            default:
                throw new Exception(
                        "Parameter \"autoscale\" must be true or false");
            }

            int nchars = Util.toInt(args[i++]);

	    GSTics tics = new GSTics(min, max, log, autoscale, nchars);
	    System.err.println(">> " + tics);
	}
}

	    
